/*

DDS / Receiver controller for JUMA RX1 Direct conversion Receiver. 
OH2NLT Juha Niinikoski 12.01.2005

Processor is Microchip 16F819 providing 2k words Flash, 256 bytes RAM and 256 bytes EEPROM.
Project compiled under MPLAB7.00 with Hitec PICC 8.05PL2 c-compiler


This code can be used freely for private Ham Radio projects.
Code is not a sample of clean good c language programming.
Limited amount of 16F819 Flash memory have forced to use
not clean but code size effective programming methods.


DDS driver aadded, initial display logic 14.01.2005

Code cleaning & size optimization 18.01.2005 99% Flash used v0.01
New compacted display conversion routines & EEPROM frequency storage 19.01.2005 78%Flash used v0.02
Freq set at init, tune digit select corrected and Pulse Encoder logic. 20.01.2005 v0.03
Appearance corrections, init sequence corrections, 500Hz tuning step, disp rounding, morse test 
86,1% Flash space used 26.01.2005 v0.04
"RF" play with 20MHz sub harmonics morse message 27.01.2005 v0.05
Encoder debounce filter, step change beep logic, RA0 audio out. 89,1% Flash used 03.02.2005 v0.06
Beep added when entering tune step change mode, code size optimization 04.02.2005 v0.07
Code size optimization, tune step save to EEPROM,  long push save logic. 90,5%  07.02.2005 v0.08
Saved text added to save logic 91,1% 10.02.2005 v0.09
Code size & other addjustments, different high limit for test mode 92,5% 13.02.2005 v0.10

Start up message display time changed, first release version 24.02.2005 v1.00
TX 700Hz CW shift logic added 21.06.2005  v2.00
Encoder clear after startup, 100Hz tuning step 15.08.2005 v2.01

PIC16F88 added as alternative processor 28.08.2005 88v201
*/

// Define this for pic 16F88 processor
// Comment out for 16F819 processor
#define F88


//**********************
// Include header files
//**********************

#include 	<pic.h>
#include	"jumarx1.h"


//****************
// Hello messages
//*****************

// Note ! Display organization is 2x8 characters in one line

const char ver[] = {"OH2NLT 28.08.2005"};		// Version info
//const char lcd_ver1[] = {" OH7SV  "}; 
//const char lcd_ver2[] = {" OH2NLT "};


//******************
// Global variables
//******************

signed char tune_digit;				// which digit to tune
signed char old_tune_digit;
#define DEF_STEP 	2				// default / start up tune step

bit pwchange;						// encoder moved
signed char pwctr;					// encoder delta
unsigned char portb_copy;			// PORTB read data from irq routine. Update intervall 1ms
			
bit dmode;							// LCD display mode, only two alternatives
bit testmode;						// test mode indicator

int ms_timer;						// millisecond timer, decremented by irq routine, stop at 0

union								// Battery voltage from A/D converter
	{
	int battery;					// A/D counts = U/2
	unsigned char adcregs[2];
	} ad;

unsigned char smeter;				// S-meter data from A/D converter. Smax = 240mV

bank1 long rx_freq;					// active frequency
#define	dds_freq		(m1.m1l)	// define storage at large mul structures
//long dds_freq;						// frequency in use
long tune_step;						// tuning step

// tuning limits
#define TUNE_LOW_LIMIT	0			// test setup can go down to 0 Hz
#define TUNE_HIGH_LIMIT 0x6666666	// 8 MHz upper limit
#define TUNE_TESTH_LIMIT 132875550	// 9,9 MHz upper limit

// tuning steps
#define TUNE_STEP_LOW 	134			// 10 Hz
#define TUNE_STEP_MID	1342		// 100 Hz
//#define TUNE_STEP_MID	6711		// 500 Hz
#define TUNE_STEP_UP	13422		// 1 kHz
#define TUNE_STEP_HI	1342177		// 100 kHz

#define TX_SHIFT		9395		// 700 Hz, when TX subtract this

// user interface timers
#define SAVE_DELAY		2000		// button press time (ms) for save frequency

// some DDS constants
#define BITS28		268435456UL		// 2^28 constant
#define BITS32		4294967296UL	// 2^32 constant
#define DDC_CLK		20000000UL		// DDS Chip clock

#define DMULT		320000000UL		// display multiplier, 0,07450581Hz/bit * 2^32

#define START_FREQ	0x2F58E21		// factory default (empty EEPROM) frequency, 3.699.000 Hz
#define MHz_cal		0xCCCCCC		// exact 1.000.000 Hz setup for calibration

//*****************
// Include C files
//*****************

#include	"ad9833.c"
#include	"large_mul.c"
#include	"dcrxtim.c"
#include	"dcrxlcd.c"
#include 	"dcrxadc.c"


//******************
// Processor config
//******************

#ifdef F88
	__CONFIG(UNPROTECT & HS & WDTDIS & PWRTEN & BOREN & LVPDIS & MCLRDIS);	// 16F88 config
//	__CONFIG(2, FCMDIS & IESODIS);
// CONFIG2 does not work with 805PL2 compiler, patch manually
#else
	__CONFIG(UNPROTECT & HS & WDTDIS & PWRTEN & BOREN & LVPDIS & MCLRDIS);	// 16F819 config
#endif
	__IDLOC(0007);				// Device ID


//******************
// Helper functions
//******************


// generate tone RA0 pin (battery voltage measurement)

void tone(int pitch, int len)		// pitch = 1/2 cycle in us units, len = ms
	{
	ms_timer = len;
	while(ms_timer != 0)
		{
		TONE_OUT = 1;
		delayu(pitch);
		TONE_OUT = 0;
		delayu(pitch);
		}
	}

// tone definitions

#define LOW_TONE	1911	// C 261,6 Hz
#define MID_TONE	1517	// E 329,6 Hz
#define HIGH_TONE	1276	// G 392 Hz
#define HIGH2_TONE	1012	// B 493,8 Hz

#define CHANGE_LEN	100		// step change beep length (ms)


#define MORSE_TONE	955		// 523,2 Hz
#define DI 0
#define DAA 1
#define UNIT 40				// morse unit length (ms)


// generate morse symbol

void beep(char l)
	{
	if(l == DI)
		tone(MORSE_TONE, UNIT);
	else
		tone(MORSE_TONE, (3*UNIT));
	delaym(UNIT);
	}

// say CW hello 

void morse_message(void)		// hello experiment
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say JUMA
	beep(DAA);
	beep(DAA);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DAA);	
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DAA);
	delaym(7*UNIT);				// word space

	beep(DI);					// say RX1
	beep(DAA);
	beep(DI);
	delaym(3*UNIT);
	beep(DAA);
	beep(DI);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);	
	beep(DAA);
	beep(DAA);
	beep(DAA);
	beep(DAA);
	delaym(3*UNIT);

#ifdef F88
	beep(DAA);					// we have more memory, say also QRV
	beep(DAA);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DAA);
	beep(DI);
	delaym(3*UNIT);
	beep(DI);
	beep(DI);
	beep(DI);
	beep(DAA);
#endif

	TONE_DIR = 1;				// return to input mode
//	delaym(7*UNIT);				// end delay
	}

void morse_roger(void)				// say roger
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say R
	beep(DAA);
	beep(DI);

	TONE_DIR = 1;				// return to input mode
	}

// this loop generates 625kHz signal J1-1 ( RB4 )
// from 20MHz clock

#if 0
void meas_osc(void)				// No return Oscillator measurement loop
	{
	TEST_OUT_TS = 0;			// make J1-1 output for Oscillator test
loop:
	TEST_OUT = 1;
	asm("nop");
	asm("nop");
	asm("nop");

	TEST_OUT = 0;
	goto loop;
	}
#endif


// store & read frequency to/from internal EEPROM
// EEPROM[0...3] = frequency, EEPROM[4] = step
void save_freq(void)
	{
	char x;
	for(x=0; x<4; x++)
		{
		EEPROM_WRITE(x, (m1.m1b[x]));		// save frequency in use (DDS tuning word)
		}
	EEPROM_WRITE(4, tune_digit);			// save tune digit
	}

void read_freq(void)
	{
	char x;
	for(x=0; x<4; x++)
		{
		m1.m1b[x] = EEPROM_READ(x);			// restore directly to large mul structure
		}
	tune_digit = EEPROM_READ(4);			// restore tuning step
	}


// check rx_frequency (DDS tuning word) limits and
// force rx_freq into limits
// status = 0 if ok, 1 if low, 2 if high

char check_freq_limits(void)
	{
	if((long)dds_freq < TUNE_LOW_LIMIT)		// limitters
		{
		dds_freq = TUNE_LOW_LIMIT;			// low = 0
		return(1);
		}
	if(testmode)
		{
		if(dds_freq > TUNE_TESTH_LIMIT)
			{
			dds_freq = TUNE_TESTH_LIMIT;			// high = 9,9 MHz
			return(2);
			}
		}
	else
		{
		if(dds_freq > TUNE_HIGH_LIMIT)
			{
			dds_freq = TUNE_HIGH_LIMIT;			// high = 8 MHz
			return(2);
			}
		}

	return(0);
	}

// Generate beep tone according tune step 
#ifdef F88
const int tune_tone[5] = {0, LOW_TONE, MID_TONE, HIGH_TONE, HIGH2_TONE};	// 16F88 tune step tone indicators
#else
const int tune_tone[4] = {0, LOW_TONE, MID_TONE, HIGH_TONE};	// tune step tone indicators
#endif

void step_beep( void )
	{
	TONE_DIR = 0;								// make tone output active

	tone(tune_tone[tune_digit], CHANGE_LEN);	// select tone & beep

	TONE_DIR = 1;								// return to input mode
	} 


//*****************
// Main function
//*****************

void main(void)
{
/* Init IO Ports & software modules */

//Set up ADC and PORTs

#ifdef F88
	ANSEL = ANSEL_INIT;		// A/D converter pin select
#endif

	PORTA = PORTA_INIT;		// A/D & control signals
	PORTB = PORTB_INIT;		// LCD & SW port

	ADCON1 = ADCON1_INIT;	// PORTA = digital I/O during start

	TRISA = TRISA_INIT;		// set port direction bits
	TRISB = TRISB_INIT;

	Init_timers();			// Init timers & interrupt system

	init_ad9833();			// init DDS

	read_freq();			// read start up value from EEPROM & store it to dds_freq
	if(check_freq_limits() != 0)			// check limits, empty EEPROM ?
		{
		dds_freq = START_FREQ;				// if check fails use cold start definition
		tune_digit = DEF_STEP;
		}
	old_tune_digit = tune_digit;
	tune_ad9833(dds_freq);	// set DDS
	rx_freq = dds_freq;		// set rx_freq

	lcdinit();				// Init LCD

	putst("JUMA-RX1");		// Show RX type & version
	lcdcom(SECOND_HALF);
#ifdef F88
	putst(" v201/88");
#else
	putst("   v2.01");		// special 16F88 version message
#endif
	delaym(200);			// wait for audio ready
	pwctr = 0;				// clear encoder
	morse_message();		// say audible hello

	lcdclr( );
	putst(" OH7SV  ");		// Tell who we are
	lcdcom(SECOND_HALF);
	putst(" OH2NLT ");
	delaym(1800);			// "optimized" message display time

	set_chgen();			// download soft characters
	lcdclr( );				// clear display


	if((portb_copy & BUTTON_1) == 0) // Oscillator test ?
		{
		putst("1 MHz");		// show calibration message
		tune_ad9833(MHz_cal);	// set 1.000.000 MHz
		rx_freq = MHz_cal;
		delaym(10000);		// show message 10s
		testmode = 1;		// set test mode flag
		}

// Free running Main Control Loop
// Loop time: About 5ms when voltage display selected
// About 6,3ms when S-meter display selected
// 1ms button pressed.

	for(;;)
	{

// do A/D measurements

	measure_all();								// measure & store A/D channells


// addjust display

	if((portb_copy & BUTTON_1) == 0) 			// test encoder button function
		{
		lcdclr();
		pwctr = 0;
		pwchange = 0;							// clear encoder for setup function
		step_beep();							// entry generate beep, tell where we are
		ms_timer = SAVE_DELAY;					// arm save timer, ! note step_beep() use same timer !

		while((portb_copy & BUTTON_1) == 0)		// display & tune setup loop
			{
			tune_digit = tune_digit + pwctr;
			if(tune_digit<1)
				tune_digit = 1;					// check limits(1...3) & stay between them
#ifdef F88
			if(tune_digit>4)					// 16F88 version have one more
				tune_digit = 4;
#else
			if(tune_digit>3)
				tune_digit = 3;
#endif
			pwctr = 0;

			lcdcom(FIRST_HALF);
			if(testmode)						// test mode display
				putlonghex(rx_freq);			// show DDS control word

			lcdcom(SECOND_HALF);							// show which digit to tune
			putst("  Step ");
			puthexnible(tune_digit);

			if(tune_digit != old_tune_digit)	// generate beep if step changed
				{
				step_beep();					// generate beep
				old_tune_digit = tune_digit;	// save change
				}

			if(ms_timer == 0 && pwchange == 0)	// long press && encoder not moved
				{
				pwchange = 1;					// save only once per push
				lcdcom(SECOND_HALF);			// set cursor to second half
				putst("  Saved ");				// message to first half
				save_freq();					// save frequency to EEPROM
				morse_roger();					// roger
				while((portb_copy & BUTTON_1) == 0) // wait here for button release
					;
				}

			}

		if(pwchange == 1)						// encoder turned ?
			{
			pwchange = 0;						//reset flag, do nothing more
			}
		else
			{
			dmode = !dmode;						// change display mode
			}
		}


// tune RX

	if(pwchange == 1)							// do encoder, addjust frequency
		{
#ifdef F88
		switch(tune_digit)						// 16F88 version of tuning steps
			{
			case 1: default:
			tune_step = TUNE_STEP_LOW;			// fine tuning
			break;

			case 2:
			tune_step = TUNE_STEP_MID;			//  mid range tuning
			break;

			case 3:
			tune_step = TUNE_STEP_UP;			//  mid/high range tuning
			break;

			case 4:
			tune_step = TUNE_STEP_HI;			//  coarse tuning
			break;
			}
#else
		if(tune_digit == 1)						// set tuning step, with save memory methode
			tune_step = TUNE_STEP_LOW;			// fine tuning
		else
			{
			if(tune_digit == 2)					// step test 26.01.2005

				tune_step = TUNE_STEP_MID;		// mid range tuning
			else
				tune_step = TUNE_STEP_HI;		// coarse tuning
			}
#endif
		rx_freq = rx_freq + pwctr * tune_step;	// add +/- steps

		pwchange = 0;							// clear encoder
		pwctr = 0;
		}


// calculate shift if TX on

	if((portb_copy & TX) == 0)
		dds_freq = rx_freq - TX_SHIFT;	// subtract tx shift
	else
		dds_freq = rx_freq;

		check_freq_limits();				// check limits (dds_freq)
		tune_ad9833(dds_freq);				// set DDS


// Format & Display selected page

	lcdcom(FIRST_HALF);						// select line one (first 8 chr) display
	if(dmode == 0)							// Display page 0
		{
		draw_s_meter(smeter);				// draw smeter graph, scale = 0...48 counts
		}
	else									// Display page 1
		{
		putvolt(ad.battery << 1);			// A/D count * 2 = voltage  12.00V
		putst(" V ");
		}
	
	lcdcom(SECOND_HALF);					// line two allways frequency display
//	m1.m1l = dds_freq;						// calculate display frequency from DDS setup word
	m2.m2l = DMULT;
	vlmul();								// very large multiply, 32*32 bits
	add_prod(3, 0x80);						// round, add 2^32/2 to product before truncation 26.01.2005
	putfreq(prod.prodlng.prodhi);			// display frequency
	}
}
